"""This file defines all the functions/classes to perform any CRUD operation on
the Role table, for a given domain node, in a RESTful manner.
"""

# stdlib
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.signing import VerifyKey

# relative
from .....common.message import ImmediateSyftMessageWithReply
from ....domain.domain_interface import DomainInterface
from ...exceptions import AuthorizationError
from ...exceptions import MissingRequestKeyError
from ...exceptions import RequestError
from ...exceptions import RoleNotFoundError
from ...node_table.utils import model_to_json
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..success_resp_message import SuccessResponseMessage
from .role_manager_messages import CreateRoleMessage
from .role_manager_messages import DeleteRoleMessage
from .role_manager_messages import GetRoleMessage
from .role_manager_messages import GetRoleResponse
from .role_manager_messages import GetRolesMessage
from .role_manager_messages import GetRolesResponse
from .role_manager_messages import UpdateRoleMessage

INPUT_TYPE = Union[
    Type[CreateRoleMessage],
    Type[UpdateRoleMessage],
    Type[GetRoleMessage],
    Type[GetRolesMessage],
    Type[DeleteRoleMessage],
]

INPUT_MESSAGES = Union[
    CreateRoleMessage,
    UpdateRoleMessage,
    GetRoleMessage,
    GetRolesMessage,
    DeleteRoleMessage,
]

OUTPUT_MESSAGES = Union[SuccessResponseMessage, GetRoleResponse, GetRolesResponse]


def create_role_msg(
    msg: CreateRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    """Creates a new role in the database.

    Args:
        msg (CreateRoleMessage): details of the role.
        node (DomainInterface): domain node.
        verify_key (VerifyKey): public digital signature/key of the user.

    Raises:
        MissingRequestKeyError: If name of the role is missing.
        RequestError: If role already exists.
        AuthorizationError: If user does not have permissions to create new role.

    Returns:
        SuccessResponseMessage: Success message on role creation.
    """
    # Check if user has permissions to create new roles
    _allowed = node.users.can_edit_roles(verify_key=verify_key)

    if not msg.name:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (name)!"
        )

    # Check if this role name was already registered
    try:
        node.roles.first(name=msg.name)
        raise RequestError(message="The role name already exists!")
    except RoleNotFoundError:
        pass

    if _allowed:
        node.roles.register(
            name=msg.name,
            can_make_data_requests=msg.can_make_data_requests,
            can_triage_data_requests=msg.can_triage_data_requests,
            can_manage_privacy_budget=msg.can_manage_privacy_budget,
            can_create_users=msg.can_create_users,
            can_manage_users=msg.can_manage_users,
            can_edit_roles=msg.can_edit_roles,
            can_manage_infrastructure=msg.can_manage_infrastructure,
            can_upload_data=msg.can_upload_data,
            can_upload_legal_document=msg.can_upload_legal_document,
            can_edit_domain_settings=msg.can_edit_domain_settings,
        )
    else:
        raise AuthorizationError("You're not allowed to create a new Role!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Role created successfully!",
    )


def update_role_msg(
    msg: UpdateRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    """Updates the properties of the given role.

    Args:
        msg (UpdateRoleMessage): stores msg address and properties of the role to be updated.
        node (DomainInterface): domain node.
        verify_key (VerifyKey): public digital signature of the user.

    Raises:
        MissingRequestKeyError: If the role id does not exist in the `msg`.
        AuthorizationError: If user does not have permissions to perform the update operation.

    Returns:
        SuccessResponseMessage: Message on successfully updating the role.
    """

    params = {
        "name": msg.name,
        "can_make_data_requests": msg.can_make_data_requests,
        "can_triage_data_requests": msg.can_triage_data_requests,
        "can_manage_privacy_budget": msg.can_manage_privacy_budget,
        "can_create_users": msg.can_create_users,
        "can_manage_users": msg.can_manage_users,
        "can_edit_roles": msg.can_edit_roles,
        "can_manage_infrastructure": msg.can_manage_infrastructure,
        "can_upload_data": msg.can_upload_data,
        "can_upload_legal_document": msg.can_upload_legal_document,
        "can_edit_domain_settings": msg.can_edit_domain_settings,
    }

    if not msg.role_id:
        raise MissingRequestKeyError

    # Check if user has permissions to edit roles
    _allowed = node.users.can_edit_roles(verify_key=verify_key)

    if _allowed:
        node.roles.set(role_id=msg.role_id, params=params)
    else:
        raise AuthorizationError("You're not authorized to edit this role!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Role updated successfully!",
    )


def get_role_msg(
    msg: GetRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetRoleResponse:
    """Retrieves details of a role.

    Args:
        msg (GetRoleMessage): stores msg address and role id.
        node (DomainInterface): domain node.
        verify_key (VerifyKey): public digital signature of the user.

    Raises:
        AuthorizationError: If user does not have permissions to get role information.

    Returns:
        GetRoleResponse: details of the role.
    """

    # Check if user has permissions to triage requests
    _allowed = node.users.can_triage_requests(verify_key=verify_key)

    if _allowed:
        role = node.roles.first(id=msg.role_id)
        _msg = model_to_json(role)
    else:
        raise AuthorizationError(
            "get_role_msg You're not allowed to get User information!"
        )

    return GetRoleResponse(address=msg.reply_to, content=_msg)


def get_all_roles_msg(
    msg: GetRolesMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetRolesResponse:
    """Retrieves details of the all available roles.

    Args:
        msg (GetRolesMessage): stores the address of the message.
        node (DomainInterface): domain node.
        verify_key (VerifyKey): public digital signature of the user.

    Raises:
        AuthorizationError: If user does not have permissions to access roles.

    Returns:
        GetRolesResponse: stores the details for all the roles as a list.
    """

    # Check if user has permissions to view roles
    _allowed = node.users.can_triage_requests(verify_key=verify_key)

    if _allowed:
        roles = node.roles.all()
        _msg = [model_to_json(role) for role in roles]
    else:
        raise AuthorizationError("You're not allowed to get Role information!")

    return GetRolesResponse(address=msg.reply_to, content=_msg)


def del_role_msg(
    msg: DeleteRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    """Delete the role corresponding to the given role id.

    Args:
        msg (DeleteRoleMessage): stores the msg address and id of the role to be deleted.
        node (DomainInterface): domain node.
        verify_key (VerifyKey): public digital signature of the user.

    Raises:
        AuthorizationError: If user does not have permissions to edit roles.

    Returns:
        SuccessResponseMessage: stores the response msg on successful role deletion.
    """

    # Check if user has permissions to edit roles
    _allowed = node.users.can_edit_roles(verify_key=verify_key)

    if _allowed:
        node.roles.delete(id=msg.role_id)
    else:
        raise AuthorizationError("You're not authorized to delete this role!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Role has been deleted!",
    )


class RoleManagerService(ImmediateNodeServiceWithReply):
    """A class to handle all operations performed on the Role table."""

    msg_handler_map: Dict[INPUT_TYPE, Callable[..., OUTPUT_MESSAGES]] = {
        CreateRoleMessage: create_role_msg,
        UpdateRoleMessage: update_role_msg,
        GetRoleMessage: get_role_msg,
        GetRolesMessage: get_all_roles_msg,
        DeleteRoleMessage: del_role_msg,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: INPUT_MESSAGES,
        verify_key: VerifyKey,
    ) -> OUTPUT_MESSAGES:
        return RoleManagerService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            CreateRoleMessage,
            UpdateRoleMessage,
            GetRoleMessage,
            GetRolesMessage,
            DeleteRoleMessage,
        ]



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from .....proto.core.node.common.action.smpc_action_message_pb2 import (
    SMPCActionMessage as SMPCActionMessage_PB,
)
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address


@serializable()
class SMPCActionMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        name_action: str,
        self_id: UID,
        args_id: List[UID],
        kwargs_id: Dict[str, UID],
        result_id: UID,
        address: Address,
        kwargs: Optional[Dict[str, Any]] = None,
        ranks_to_run_action: Optional[List[int]] = None,
        msg_id: Optional[UID] = None,
    ) -> None:
        self.name_action = name_action
        self.self_id = self_id
        self.args_id = args_id
        self.kwargs_id = kwargs_id
        if kwargs is None:
            self.kwargs = {}
        else:
            self.kwargs = kwargs
        self.id_at_location = result_id
        self.ranks_to_run_action = ranks_to_run_action if ranks_to_run_action else []
        super().__init__(address=address, msg_id=msg_id)

    @staticmethod
    def filter_actions_after_rank(
        rank: int, actions: List[SMPCActionMessage]
    ) -> List[SMPCActionMessage]:
        """
        Filter the actions depending on the rank of each party

        Arguments:
            rank (int): the rank of the party
            actions (List[SMPCActionMessage]):

        """
        res_actions = []
        for action in actions:
            if rank in action.ranks_to_run_action:
                res_actions.append(action)

        return res_actions

    def __str__(self) -> str:
        res = f"SMPCAction: {self.name_action}, "
        res = f"{res}Self ID: {self.self_id}, "
        res = f"{res}Args IDs: {self.args_id}, "
        res = f"{res}Kwargs IDs: {self.kwargs_id}, "
        res = f"{res}Kwargs : {self.kwargs}, "
        res = f"{res}Result ID: {self.id_at_location}, "
        res = f"{res}Ranks to run action: {self.ranks_to_run_action}"
        return res

    __repr__ = __str__

    def _object2proto(self) -> SMPCActionMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: SMPCActionMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return SMPCActionMessage_PB(
            name_action=self.name_action,
            self_id=sy.serialize(self.self_id),
            args_id=list(map(lambda x: sy.serialize(x), self.args_id)),
            kwargs_id={k: sy.serialize(v) for k, v in self.kwargs_id.items()},
            kwargs={k: sy.serialize(v, to_bytes=True) for k, v in self.kwargs.items()},
            id_at_location=sy.serialize(self.id_at_location),
            address=sy.serialize(self.address),
            msg_id=sy.serialize(self.id),
        )

    @staticmethod
    def _proto2object(proto: SMPCActionMessage_PB) -> SMPCActionMessage:
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of SMPCActionMessage
        :rtype: SMPCActionMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return SMPCActionMessage(
            name_action=proto.name_action,
            self_id=sy.deserialize(blob=proto.self_id),
            args_id=list(map(lambda x: sy.deserialize(blob=x), proto.args_id)),
            kwargs_id={k: sy.deserialize(blob=v) for k, v in proto.kwargs_id.items()},
            kwargs={
                k: sy.deserialize(blob=v, from_bytes=True)
                for k, v in proto.kwargs.items()
            },
            result_id=sy.deserialize(blob=proto.id_at_location),
            address=sy.deserialize(blob=proto.address),
            msg_id=sy.deserialize(blob=proto.msg_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return SMPCActionMessage_PB



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......util import traceback_and_raise
from ....abstract.node_service_interface import NodeServiceInterface
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from .ping_messages import PingMessage
from .ping_messages import PingMessageWithReply
from .ping_messages import PingReplyMessage


class PingService(ImmediateNodeServiceWithReply):

    # TODO: Change this to not be available for guests
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: NodeServiceInterface,
        msg: PingMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> PingReplyMessage:
        if verify_key is None:
            traceback_and_raise("Can't process PingService with no verification key.")

        result = msg.payload.run(node=node, verify_key=verify_key)
        return PingMessageWithReply(kwargs=result).back_to(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[PingMessage]]:
        return [PingMessage]



# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ......logger import debug
from ......proto.core.node.common.service.child_node_lifecycle_service_pb2 import (
    RegisterChildNodeMessage as RegisterChildNodeMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
class RegisterChildNodeMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        lookup_id: UID,
        child_node_client_address: Address,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.lookup_id = lookup_id
        self.child_node_client_address = child_node_client_address

    def _object2proto(self) -> RegisterChildNodeMessage_PB:
        debug(f"> {self.icon} -> Proto 🔢")
        return RegisterChildNodeMessage_PB(
            lookup_id=sy.serialize(
                self.lookup_id
            ),  # TODO: not sure if this is needed anymore
            child_node_client_address=sy.serialize(self.child_node_client_address),
            address=sy.serialize(self.address),
            msg_id=sy.serialize(self.id),
        )

    @staticmethod
    def _proto2object(proto: RegisterChildNodeMessage_PB) -> "RegisterChildNodeMessage":
        msg = RegisterChildNodeMessage(
            lookup_id=sy.deserialize(blob=proto.lookup_id),
            child_node_client_address=sy.deserialize(
                blob=proto.child_node_client_address
            ),
            address=sy.deserialize(blob=proto.address),
            msg_id=sy.deserialize(blob=proto.msg_id),
        )
        debug(f"> {msg.icon} <- 🔢 Proto")
        return msg

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RegisterChildNodeMessage_PB



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Optional
from typing import Tuple

# third party
import numpy as np

# relative
from ..common.serde.serializable import serializable
from .passthrough import is_acceptable_simple_type  # type: ignore


@serializable(recursive_serde=True)
class lazyrepeatarray:
    """
    When data is repeated along one or more dimensions, store it using lazyrepeatarray
    so that you can save on RAM and CPU when computing with it. Think like the opposite
    of np.broadcast, repeated values along an axis are collapsed but the .shape
    attribute of the higher dimensional projection is retained for operations.
    """

    __attr_allowlist__ = ["data", "shape"]

    def __init__(self, data: np.ndarray, shape: Tuple[int, ...]) -> None:
        """
        data: the raw data values without repeats
        shape: the shape of 'data' if repeats were included
        """

        # NOTE: all additional arguments are assumed to be broadcast if dims are shorter
        # than that of data. Example: if data.shape == (2,3,4) and
        # min_vals.shape == (2,3), then it's assumed that the full min_vals.shape is
        # actually (2,3,4) where the last dim is simply copied.
        # Example2: if data.shape == (2,3,4) and min_vals.shape == (2,1,4), then the
        # middle dimension is supposed to be copied to be min_vals.shape == (2,3,4)
        # if necessary. This is just to keep the memory footprint (and computation)
        # as small as possible.

        if isinstance(data, (bool, int, float)):
            data = np.array(data)

        # verify broadcasting works on shapes
        np.broadcast_shapes(data.shape, shape)

        self.data = data
        self.shape = shape

    def __add__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            return self.__class__(data=self.data + other, shape=self.shape)

        if self.shape != other.shape:
            raise Exception("cannot add tensors with different shapes")

        if self.data.shape == other.data.shape:
            return self.__class__(data=self.data + other.data, shape=self.shape)

        raise Exception("not sure how to do this yet")

    def __sub__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            return self.__class__(data=self.data - other, shape=self.shape)

        if self.shape != other.shape:
            raise Exception("cannot subtract tensors with different shapes")

        if self.data.shape == other.data.shape:
            return self.__class__(data=self.data - other.data, shape=self.shape)

        raise Exception("not sure how to do this yet")

    def __mul__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            return self.__class__(data=self.data * other, shape=self.shape)

        if self.shape != other.shape:
            raise Exception("cannot multiply tensors with different shapes")

        if self.data.shape == other.data.shape:
            return self.__class__(data=self.data * other.data, shape=self.shape)

        raise Exception("not sure how to do this yet")

    def __pow__(self, exponent: int) -> lazyrepeatarray:
        if exponent == 2:
            return self * self
        raise Exception("not sure how to do this yet")

    def copy(self, order: Optional[str] = "K") -> lazyrepeatarray:
        return self.__class__(data=np.copy(self.data, order=order), shape=self.shape)

    @property
    def size(self) -> int:
        return np.prod(self.shape)

    def sum(self, axis: Optional[int] = None) -> np.ndarray:
        if axis is None:
            # TODO: make fast
            return self.to_numpy().sum()
        else:
            raise Exception("not sure how to do this yet")

    def __eq__(self, other: Any) -> lazyrepeatarray:  # type: ignore
        if isinstance(other, lazyrepeatarray):
            if self.shape == other.shape:
                return lazyrepeatarray(data=self.data == other.data, shape=self.shape)
            else:
                result = (self.to_numpy() == other.to_numpy()).all()
                return lazyrepeatarray(data=np.array([result]), shape=result.shape)
        if isinstance(other, np.ndarray):
            try:
                _ = np.broadcast_shapes(self.shape, other.shape)
                result = (self.to_numpy() == other).all()
                return lazyrepeatarray(data=np.array([result]), shape=other.shape)
            except Exception as e:
                print(
                    "Failed to compare lazyrepeatarray with "
                    + f"{self.shape} == {other.shape} to numpy by broadcasting. {e}"
                )
                raise e

        return self == other

    @property
    def dtype(self) -> np.dtype:
        return self.data.dtype

    def astype(self, np_type: np.dtype) -> lazyrepeatarray:
        return self.__class__(self.data.astype(np_type), self.shape)

    def to_numpy(self) -> np.ndarray:
        return np.broadcast_to(self.data, self.shape)

    def __repr__(self) -> str:
        return f"<lazyrepeatarray data: {self.data} -> shape: {self.shape}>"

    def __bool__(self) -> bool:
        return self.data.__bool__()

    def all(self) -> bool:
        return self.data.all()

    def any(self) -> bool:
        return self.data.any()



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Union

# third party
from nacl.signing import VerifyKey
import numpy as np
from sympy.ntheory.factor_ import factorint

# relative
from ....core.adp.entity import DataSubjectGroup
from ....core.adp.entity import Entity
from ....core.adp.scalar.intermediate_gamma_scalar import IntermediateGammaScalar
from ....util import concurrency_count
from ....util import list_sum
from ....util import parallel_execution
from ....util import split_rows
from ...adp.publish import publish
from ...adp.vm_private_scalar_manager import VirtualMachinePrivateScalarManager
from ...common.serde.serializable import serializable
from ...tensor.passthrough import PassthroughTensor  # type: ignore
from ...tensor.passthrough import is_acceptable_simple_type  # type: ignore
from ..broadcastable import is_broadcastable
from .adp_tensor import ADPTensor

SupportedChainType = Union[int, bool, float, np.ndarray, PassthroughTensor]


@serializable(recursive_serde=True)
class IntermediateGammaTensor(PassthroughTensor, ADPTensor):

    """Functionality for tracking differential privacy when individual values
    are contributed to by multiple entities. IntermediateGammaTensor differs
    from IniitalGammaTensor only in that InitialGammaTensor has additional
    functionality in its constructor essential to when one initially begins
    tracking metadata across mutliple entities, whereas IntermediateGammaTensor
    has a simpler constructor for use when performing operations across one or
    more IntermediateGammaTensor objects.
    """

    __attr_allowlist__ = [
        "term_tensor",
        "coeff_tensor",
        "bias_tensor",
        "value_tensor",
        "scalar_manager",
        "child",
        "unique_entities",
        "n_entities",
    ]

    def __init__(
        self,
        term_tensor: np.ndarray,
        coeff_tensor: np.ndarray,
        bias_tensor: np.ndarray,
        value_tensor: Optional[np.ndarray] = None,
        # min_vals: np.ndarray,
        # max_vals: np.ndarray,
        scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None,
        unique_entities: Optional[Set[Entity]] = None,
    ) -> None:
        super().__init__(term_tensor)

        # EXPLAIN A: if our clipped polynomial is y = clip(mx + b, min=min_vals, max=max_vals)
        # EXPLAIN B: if self.child = 5x10

        # EXPLAIN A: this is "x"
        # EXPLAIN B: this is a 5x10x1
        self.term_tensor = term_tensor

        # EXPLAIN A: this is "m"
        # EXPLAIN B: this is a 5x10x1
        self.coeff_tensor = coeff_tensor

        # EXPLAIN A: this is "b"
        # EXPLAIN B: this is a 5x10
        self.bias_tensor = bias_tensor

        # optional cache of values
        self.value_tensor = value_tensor

        # EXPLAIN A: this is "min_vals"
        # EXPLAIN B: this is a 5x10
        # self.min_vals = min_vals

        # EXPLAIN A: this is "max_vals"
        # EXPLAIN B: this is a 5x10
        # self.max_vals = max_vals

        self.scalar_manager = (
            scalar_manager
            if scalar_manager is not None
            else VirtualMachinePrivateScalarManager()
        )

        if not hasattr(self, "_min_vals_cache"):
            self._min_vals_cache: Optional[np.array] = None
        if not hasattr(self, "_max_vals_cache"):
            self._max_vals_cache: Optional[np.array] = None

        self.unique_entities: set[Entity] = (
            unique_entities if unique_entities is not None else set()
        )
        self.n_entities: int = len(self.unique_entities)

        if self.n_entities == 0:
            # Unique entities

            for entity in set(self._entities_list()):
                if isinstance(entity, Entity):
                    if entity not in self.unique_entities:
                        self.unique_entities.add(entity)
                        self.n_entities += 1
                elif isinstance(entity, DataSubjectGroup):
                    for e in entity.entity_set:
                        if e not in self.unique_entities:
                            self.unique_entities.add(e)
                            self.n_entities += 1
                else:
                    raise Exception(f"{type(entity)}")

    @property
    def flat_scalars(self) -> List[Any]:
        flattened_terms = self.term_tensor.reshape(-1, self.term_tensor.shape[-1])
        flattened_coeffs = self.coeff_tensor.reshape(-1, self.coeff_tensor.shape[-1])
        flattened_bias = self.bias_tensor.reshape(-1)
        # flattened_min_vals = self._min_values().reshape(-1)
        # flattened_max_vals = self._max_values().reshape(-1)

        known_primes: set[int] = set(self.scalar_manager.prime2symbol.keys())

        scalars = list()
        for i in range(len(flattened_terms)):

            single_poly_terms = flattened_terms[i]
            single_poly_coeffs = flattened_coeffs[i]
            single_poly_bias = flattened_bias[i]
            # single_poly_min_val = flattened_min_vals[i]
            # single_poly_max_val = flattened_max_vals[i]

            input_mp = [single_poly_bias]

            for j in range(len(single_poly_terms)):
                term = single_poly_terms[j]
                coeff = single_poly_coeffs[j]

                if term in known_primes:
                    prime_list = [(term, 1)]
                else:
                    prime_list = factorint(term).items()

                for prime, n_times in prime_list:
                    input_scalar = self.scalar_manager.prime2symbol[int(prime)]
                    right = input_scalar * n_times * coeff
                    input_mp.append(right)

            num_process = min(concurrency_count(), len(input_mp))
            if num_process > 1:
                args = split_rows(input_mp, cpu_count=num_process)
                output = parallel_execution(list_sum, cpu_bound=True)(args)
                scalar: IntermediateGammaScalar = sum(output)  # type: ignore
            else:
                scalar: IntermediateGammaScalar = sum(input_mp)  # type: ignore

            # to optimize down stream we can prevent search on linear queries if we
            # know that all single_poly_terms are prime therefore the query is linear
            scalar.is_linear = True
            if j in single_poly_terms:
                if (
                    single_poly_terms[j] not in known_primes
                    or single_poly_coeffs[j] != 1
                ):
                    scalar.is_linear = False
                    break

            scalars.append(scalar)

        if self.value_tensor is not None:
            flat_value_tensor = self.value_tensor.flatten()
            for i in range(len(scalars)):
                scalars[i]._value_cache = flat_value_tensor[i]  # type: ignore

        return scalars

    def update(self) -> None:
        # TODO: Recalculate term tensor using an optimized slice method

        # index_number = self._entities()  # TODO: Check if this works!!
        prime_numbers = np.array(self.scalar_manager.primes_allocated, dtype=np.int32)

        # TODO: See if this fails for addition of more than 2 IGTs
        length = len(prime_numbers)
        tensor_shape_count = np.prod(self.term_tensor.shape)
        if length == tensor_shape_count:
            self.term_tensor = prime_numbers.reshape(self.term_tensor.shape)
        elif length > tensor_shape_count:
            self.term_tensor = prime_numbers[-tensor_shape_count:].reshape(
                self.term_tensor.shape
            )
        else:
            raise Exception(f"Failed to update IGT with {length} {tensor_shape_count}")

        # self.term_tensor.flatten().reshape(-1, 2)[:, -1] = prime_numbers

    def _values(self) -> np.array:
        """WARNING: DO NOT MAKE THIS AVAILABLE TO THE POINTER!!!
        DO NOT ADD THIS METHOD TO THE AST!!!
        """
        return np.array(list(map(lambda x: x.value, self.flat_scalars))).reshape(
            self.shape
        )

    def _max_values(self) -> np.array:
        """WARNING: DO NOT MAKE THIS AVAILABLE TO THE POINTER!!!
        DO NOT ADD THIS METHOD TO THE AST!!!
        """
        if self._max_vals_cache is not None:
            return self._max_vals_cache

        return np.array(list(map(lambda x: x.max_val, self.flat_scalars))).reshape(
            self.shape
        )

    def _min_values(self) -> np.array:
        """WARNING: DO NOT MAKE THIS AVAILABLE TO THE POINTER!!!
        DO NOT ADD THIS METHOD TO THE AST!!!
        """

        if self._min_vals_cache is not None:
            return self._min_vals_cache

        return np.array(list(map(lambda x: x.min_val, self.flat_scalars))).reshape(
            self.shape
        )

    def _entities_list(self) -> list:
        """WARNING: DO NOT MAKE THIS AVAILABLE TO THE POINTER!!!
        DO NOT ADD THIS METHOD TO THE AST!!!
        """

        output_entities = []
        for flat_scalar in self.flat_scalars:
            # TODO: This will fail if the nested entity is any deeper than 2 levels- i.e. [A, [A, [A, B]]]. Recursive?
            combined_entities = DataSubjectGroup()
            for row in flat_scalar.input_entities:
                if isinstance(row, Entity) or isinstance(row, DataSubjectGroup):
                    combined_entities += row
                elif isinstance(row, list):
                    for i in row:
                        if isinstance(i, Entity) or isinstance(i, DataSubjectGroup):
                            combined_entities += i
                        else:
                            raise Exception(f"Not implemented for i of type:{type(i)}")
                else:
                    raise Exception(f"No plans for row type:{type(row)}")
            output_entities.append(combined_entities)
        return output_entities

    def _entities(self) -> np.array:
        """WARNING: DO NOT MAKE THIS AVAILABLE TO THE POINTER!!!
        DO NOT ADD THIS METHOD TO THE AST!!!
        """

        output_entities = []
        for flat_scalar in self.flat_scalars:
            # TODO: This will fail if the nested entity is any deeper than 2 levels- i.e. [A, [A, [A, B]]]. Recursive?
            combined_entities = DataSubjectGroup()
            for row in flat_scalar.input_entities:
                if isinstance(row, Entity) or isinstance(row, DataSubjectGroup):
                    combined_entities += row
                elif isinstance(row, list):
                    for i in row:
                        if isinstance(i, Entity) or isinstance(i, DataSubjectGroup):
                            combined_entities += i
                        else:
                            raise Exception(f"Not implemented for i of type:{type(i)}")
                else:
                    raise Exception(f"No plans for row type:{type(row)}")
            output_entities.append(combined_entities)
        return np.array(output_entities).reshape(self.shape)

    def astype(self, np_type: np.dtype = np.int32) -> IntermediateGammaTensor:
        return self.__class__(
            term_tensor=self.term_tensor.astype(np_type),
            coeff_tensor=self.coeff_tensor.astype(np_type),
            bias_tensor=self.bias_tensor.astype(np_type),
            scalar_manager=self.scalar_manager,
            unique_entities=self.unique_entities,
        )

    def __gt__(self, other: Union[np.ndarray, IntermediateGammaTensor]) -> Any:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals > other,
                    min_vals=np.zeros_like(vals),
                    max_vals=np.ones_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                tensor = InitialGammaTensor(
                    values=self_vals > other_vals,
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    def __lt__(self, other: Union[np.ndarray, IntermediateGammaTensor]) -> Any:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals < other,
                    min_vals=np.zeros_like(vals),
                    max_vals=np.ones_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                tensor = InitialGammaTensor(
                    values=self_vals < other_vals,
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    def __eq__(self, other: Union[np.ndarray, IntermediateGammaTensor]) -> Any:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals == other,
                    max_vals=np.ones_like(vals),
                    min_vals=np.zeros_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                # output_scalar = self.scalar_manager.copy()
                # output_scalar.combine_(other.scalar_manager)
                # other.update()
                tensor = InitialGammaTensor(
                    values=self_vals == other_vals,
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                    # scalar_manager=output_scalar
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    def __ne__(self, other: Union[np.ndarray, IntermediateGammaTensor]) -> Any:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals != other,
                    max_vals=np.ones_like(vals),
                    min_vals=np.zeros_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                # output_scalar = self.scalar_manager.copy()
                # output_scalar.combine_(other.scalar_manager)
                # other.update()
                tensor = InitialGammaTensor(
                    values=self_vals != other_vals,
                    # values= not (self_vals < other_vals) and not (self_vals > other_vals),
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                    # scalar_manager=output_scalar
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    def __ge__(self, other: Union[np.ndarray, IntermediateGammaTensor]) -> Any:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals >= other,
                    max_vals=np.ones_like(vals),
                    min_vals=np.zeros_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                tensor = InitialGammaTensor(
                    values=self_vals >= other_vals,
                    # values= not (self_vals < other_vals) and not (self_vals > other_vals),
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    def __le__(self, other: Union[np.ndarray, IntermediateGammaTensor]) -> Any:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals <= other,
                    max_vals=np.ones_like(vals),
                    min_vals=np.zeros_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                tensor = InitialGammaTensor(
                    values=self_vals <= other_vals,
                    # values= not (self_vals < other_vals) and not (self_vals > other_vals),
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.term_tensor.shape[:-1]

    @property
    def full_shape(self) -> Tuple[int, ...]:
        return self.term_tensor.shape

    def publish(self, acc: Any, sigma: float, user_key: VerifyKey) -> np.ndarray:
        print("IntermediaGammaTensor:510: TRY: publish(scalars=self.flat_scalars)")
        result = np.array(
            publish(
                scalars=self.flat_scalars,
                acc=acc,
                sigma=sigma,
                user_key=user_key,
                public_only=True,
            )
        ).reshape(self.shape)
        print("IntermediaGammaTensor:510: SUCCESS: publish(scalars=self.flat_scalars)")
        sharetensor_values = getattr(self, "sharetensor_values", None)
        if sharetensor_values is not None:
            # relative
            from ..smpc.share_tensor import ShareTensor

            result = ShareTensor(
                rank=sharetensor_values.rank,
                parties_info=sharetensor_values.parties_info,
                ring_size=sharetensor_values.ring_size,
                seed_przs=sharetensor_values.seed_przs,
                clients=sharetensor_values.clients,
                value=result,
            )
        return result

    def sum(self, axis: Optional[int] = None) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().sum(axis),
            entities=self._entities().sum(axis),
            max_vals=self._max_values().sum(axis),
            min_vals=self._min_values().sum(axis),
        )

    def prod(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().prod(axis),
            entities=self._entities().sum(
                axis
            ),  # Entities get added (combined) instead of multiplied
            max_vals=self._max_values().prod(axis),
            min_vals=self._min_values().prod(axis),
        )

    def __add__(self, other: Any) -> IntermediateGammaTensor:
        output_scalar_manager = self.scalar_manager.copy()
        # TODO: add support for SingleEntitiyPhiTensor
        # this will cause it to generate them using a more computationally intensive
        unique_entities = None

        if is_acceptable_simple_type(other):

            # EXPLAIN A: if our polynomail is y = mx + b
            # EXPLAIN B: if self.child = 5x10

            # EXPLAIN A: this is "x"
            # EXPLAIN B: this is a 5x10x1
            term_tensor = self.term_tensor

            # EXPLAIN A: this is "m"
            # EXPLAIN B: this is a 5x10x1
            coeff_tensor = self.coeff_tensor

            # EXPLAIN A: this is "b"
            # EXPLAIN B: this is a 5x10
            bias_tensor = self.bias_tensor + other

            max_vals_cache = self._max_values() + other
            min_vals_cache = self._min_values() + other
            unique_entities = self.unique_entities

        else:

            # relative
            from .dp_tensor_converter import convert_to_gamma_tensor
            from .single_entity_phi import SingleEntityPhiTensor

            if isinstance(other, SingleEntityPhiTensor):
                other = convert_to_gamma_tensor(other)

            # if self.scalar_manager != other.scalar_manager:
            # TODO: come up with a method for combining symbol factories
            # raise Exception(
            #     "Cannot add two tensors with different symbol encodings"
            # )

            output_scalar_manager.combine_(other.scalar_manager)
            other.update()  # change term tensor after combining scalar managers

            # Step 1: Concatenate
            term_tensor = np.concatenate([self.term_tensor, other.term_tensor], axis=-1)  # type: ignore
            coeff_tensor = np.concatenate(  # type: ignore
                [self.coeff_tensor, other.coeff_tensor], axis=-1
            )
            bias_tensor = self.bias_tensor + other.bias_tensor

            max_vals_cache = self._max_values() + other._max_values()
            min_vals_cache = self._min_values() + other._min_values()

            if hasattr(other, "unique_entities"):
                unique_entities = self.unique_entities.union(other.unique_entities)

        # EXPLAIN B: NEW OUTPUT becomes a 5x10x2
        # TODO: Step 2: Reduce dimensionality if possible (look for duplicates)
        result = IntermediateGammaTensor(
            term_tensor=term_tensor,
            coeff_tensor=coeff_tensor,
            bias_tensor=bias_tensor,
            scalar_manager=output_scalar_manager,
            unique_entities=unique_entities,
        )

        result._min_vals_cache = min_vals_cache
        result._max_vals_cache = max_vals_cache

        return result

    # def clip(self, a_min:int, a_max: int) -> IntermediateGammaTensor:
    #     """Clips the tensor at a certain minimum and maximum. a_min and a_max are
    #     assumed to be integers at present because IntermediateGammaTensor only
    #     operates over integer values at present."""
    #
    #     return None

    def __sub__(self, other: Any) -> IntermediateGammaTensor:

        if is_acceptable_simple_type(other):

            # EXPLAIN A: if our polynomail is y = mx + b
            # EXPLAIN B: if self.child = 5x10

            # EXPLAIN A: this is "x"
            # EXPLAIN B: this is a 5x10x1
            term_tensor = self.term_tensor

            # EXPLAIN A: this is "m"
            # EXPLAIN B: this is a 5x10x1
            coeff_tensor = self.coeff_tensor

            # EXPLAIN A: this is "b"
            # EXPLAIN B: this is a 5x10
            bias_tensor = self.bias_tensor - other

        else:
            if self.scalar_manager != other.scalar_manager:
                # TODO: come up with a method for combining symbol factories
                raise Exception(
                    "Cannot add two tensors with different symbol encodings"
                )

            # Step 1: Concatenate
            term_tensor = np.concatenate([self.term_tensor, other.term_tensor], axis=-1)  # type: ignore
            coeff_tensor = np.concatenate(  # type: ignore
                [self.coeff_tensor, other.coeff_tensor * -1], axis=-1
            )
            bias_tensor = self.bias_tensor - other.bias_tensor

        # EXPLAIN B: NEW OUTPUT becomes a 5x10x2
        # TODO: Step 2: Reduce dimensionality if possible (look for duplicates)
        return IntermediateGammaTensor(
            term_tensor=term_tensor,
            coeff_tensor=coeff_tensor,
            bias_tensor=bias_tensor,
            scalar_manager=self.scalar_manager,
        )

    def repeat(
        self, *args: List[Any], **kwargs: Dict[Any, Any]
    ) -> IntermediateGammaTensor:
        return IntermediateGammaTensor(
            term_tensor=self.term_tensor.repeat(*args, **kwargs).reshape(
                -1, self.term_tensor.shape[-1]
            ),
            coeff_tensor=self.coeff_tensor.repeat(*args, **kwargs).reshape(
                -1, self.coeff_tensor.shape[-1]
            ),
            bias_tensor=self.bias_tensor.repeat(*args, **kwargs),
            scalar_manager=self.scalar_manager,
        )

    def __mul__(self, other: Any) -> IntermediateGammaTensor:

        # EXPLAIN A: if our polynomial is y = mx
        # EXPLAIN B: self.child = 10x5

        if is_acceptable_simple_type(other):

            # this is "x"
            # this is 10x5x1
            term_tensor = self.term_tensor

            # term_tensor is prime because if I have variable "3" and variable "5" then
            # then variable "3 * 5 = 15" is CERTAIN to be the multiplication of ONLY "3" and "5"

            # this is "m"
            coeff_tensor = self.coeff_tensor * other

            bias_tensor = self.bias_tensor * other

        else:
            if self.scalar_manager != other.scalar_manager:
                # TODO: come up with a method for combining symbol factories
                raise Exception(
                    "Cannot add two tensors with different symbol encodings"
                )
            # relative
            from .dp_tensor_converter import convert_to_gamma_tensor
            from .single_entity_phi import SingleEntityPhiTensor

            if isinstance(other, SingleEntityPhiTensor):
                other = convert_to_gamma_tensor(other)

            terms = list()
            for self_dim in range(self.term_tensor.shape[-1]):
                for other_dim in range(other.term_tensor.shape[-1]):
                    new_term = np.expand_dims(  # type: ignore
                        self.term_tensor[..., self_dim]
                        * other.term_tensor[..., other_dim],
                        -1,
                    )
                    terms.append(new_term)

            for self_dim in range(self.term_tensor.shape[-1]):
                new_term = np.expand_dims(self.term_tensor[..., self_dim], -1)  # type: ignore
                terms.append(new_term)

            for _ in range(self.term_tensor.shape[-1]):
                new_term = np.expand_dims(other.term_tensor[..., self_dim], -1)  # type: ignore
                terms.append(new_term)

            term_tensor = np.concatenate(terms, axis=-1)  # type: ignore

            coeffs = list()
            for self_dim in range(self.coeff_tensor.shape[-1]):
                for other_dim in range(other.coeff_tensor.shape[-1]):
                    new_coeff = np.expand_dims(  # type: ignore
                        self.coeff_tensor[..., self_dim]
                        * other.coeff_tensor[..., other_dim],
                        -1,
                    )
                    coeffs.append(new_coeff)

            for self_dim in range(self.coeff_tensor.shape[-1]):
                new_coeff = np.expand_dims(  # type: ignore
                    self.coeff_tensor[..., self_dim] * other.bias_tensor, -1
                )
                coeffs.append(new_coeff)

            for _ in range(self.coeff_tensor.shape[-1]):
                new_coeff = np.expand_dims(  # type: ignore
                    other.coeff_tensor[..., self_dim] * self.bias_tensor, -1
                )
                coeffs.append(new_coeff)

            coeff_tensor = np.concatenate(coeffs, axis=-1)  # type: ignore

            bias_tensor = self.bias_tensor * other.bias_tensor

        # TODO: Step 2: Reduce dimensionality if possible (look for duplicates)
        return IntermediateGammaTensor(
            term_tensor=term_tensor,
            coeff_tensor=coeff_tensor,
            bias_tensor=bias_tensor,
            scalar_manager=self.scalar_manager,
        )

    def __matmul__(self, other: Any) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        self_values = self._values()
        self_entities = self._entities()
        self_min = self._min_values()
        self_max = self._max_values()

        # Private-Public
        if isinstance(other, np.ndarray):
            return InitialGammaTensor(
                values=self_values.__matmul__(other),
                entities=self_entities.__matmul__(other),
                min_vals=self_min.__matmul__(other),
                max_vals=self_max.__matmul__(other),
            )
        else:  # Private-Private
            # relative
            from .single_entity_phi import SingleEntityPhiTensor

            if isinstance(other, SingleEntityPhiTensor):
                # relative
                from .dp_tensor_converter import convert_to_gamma_tensor

                other_gamma = convert_to_gamma_tensor(other)
                return InitialGammaTensor(
                    values=self_values.__matmul__(other_gamma._values()),
                    entities=self_entities.__matmul__(other_gamma._entities()),
                    min_vals=self_min.__matmul__(other_gamma._min_values()),
                    max_vals=self_max.__matmul__(other_gamma._max_values()),
                )
            elif isinstance(other, IntermediateGammaTensor):
                return InitialGammaTensor(
                    values=self_values.__matmul__(other._values()),
                    entities=self_entities.__matmul__(other._entities()),
                    min_vals=self_min.__matmul__(other._min_values()),
                    max_vals=self_max.__matmul__(other._max_values()),
                )
            raise NotImplementedError

    def __pos__(self) -> IntermediateGammaTensor:
        return self

    def __neg__(self) -> IntermediateGammaTensor:
        return IntermediateGammaTensor(
            term_tensor=self.term_tensor,
            coeff_tensor=-self.coeff_tensor,
            bias_tensor=-self.bias_tensor,
            scalar_manager=self.scalar_manager,
        )

    def copy(self) -> IntermediateGammaTensor:
        return IntermediateGammaTensor(
            term_tensor=self.term_tensor,
            coeff_tensor=self.coeff_tensor,
            bias_tensor=self.bias_tensor,
            scalar_manager=self.scalar_manager,
        )

    def flatten(self, order: Optional[str] = "C") -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().flatten(order),
            entities=self._entities().flatten(order),
            min_vals=self._min_values().flatten(order),
            max_vals=self._max_values().flatten(order),
            scalar_manager=self.scalar_manager,
        )

    def transpose(
        self, axes: Optional[Union[int, Sequence[int], Tuple[int]]] = None
    ) -> IntermediateGammaTensor:
        # TODO: Need to check if new prime numbers are issued or if old ones are just moved around.
        # relative
        from .initial_gamma import InitialGammaTensor

        output_values = self._values().transpose(axes)
        shape = output_values.shape

        return InitialGammaTensor(
            values=output_values,
            entities=self._entities().transpose(axes),
            min_vals=self._min_values().reshape(shape),
            max_vals=self._max_values().reshape(shape),
            scalar_manager=self.scalar_manager,
        )

    def reshape(self, *shape: Sequence[int]) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        output_values = self._values().reshape(*shape)
        target_shape = output_values.shape

        return InitialGammaTensor(
            values=output_values,
            entities=self._entities().reshape(target_shape),
            min_vals=self._min_values().reshape(target_shape),
            max_vals=self._max_values().reshape(target_shape),
            scalar_manager=self.scalar_manager,
        )

    def resize(self, *new_shape: Sequence[int]) -> IntermediateGammaTensor:

        """Currently, we are making all mutable operations (like resize) return a new tensor
        instead of modifying in place.
        This currently means resize == reshape."""
        return self.reshape(*new_shape)

    #     # relative
    #     from .initial_gamma import InitialGammaTensor
    #
    #     output_values = self._values()
    #     output_values.resize(new_shape, refcheck)
    #     shape = output_values.shape
    #     output_tensor = InitialGammaTensor(
    #         values=output_values,
    #         entities=self._entities().reshape(shape),
    #         max_vals=self._max_values().reshape(shape),
    #         min_vals=self._min_values().reshape(shape),
    #         scalar_manager=self.scalar_manager,
    #     )
    #
    #     # Copy all members from the new object
    #     self.__dict__ = output_tensor.__dict__
    # self.term_tensor = output_tensor.term_tensor
    # self.coeff_tensor = output_tensor.coeff_tensor
    # self.bias_tensor = output_tensor.bias_tensor

    def ravel(self, order: Optional[str] = "C") -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().ravel(order),
            entities=self._entities().ravel(order),
            min_vals=self._min_values().ravel(order),
            max_vals=self._max_values().ravel(order),
            scalar_manager=self.scalar_manager,
        )

    def squeeze(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().squeeze(axis),
            entities=self._entities().squeeze(axis),
            min_vals=self._min_values().squeeze(axis),
            max_vals=self._max_values().squeeze(axis),
            scalar_manager=self.scalar_manager,
        )

    def swapaxes(self, axis1: int, axis2: int) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().swapaxes(axis1, axis2),
            entities=self._entities().swapaxes(axis1, axis2),
            min_vals=self._min_values().swapaxes(axis1, axis2),
            max_vals=self._max_values().swapaxes(axis1, axis2),
            scalar_manager=self.scalar_manager,
        )

    # def partition(
    #     self,
    #     kth: Union[int, Tuple[int, ...]],
    #     axis: Optional[int] = -1,
    #     kind: Optional[str] = "introselect",
    #     order: Optional[Union[int, Tuple[int, ...]]] = None,
    # ) -> IntermediateGammaTensor:
    #     # relative
    #     from .initial_gamma import InitialGammaTensor
    #
    #     return InitialGammaTensor(
    #         values=self._values().partition(kth, axis, kind, order),
    #         entities=self._entities().partition(kth, axis, kind, order),
    #         min_vals=self._min_values().partition(kth, axis, kind, order),
    #         max_vals=self._max_values().partition(kth, axis, kind, order),
    #     )

    def compress(
        self,
        condition: List[bool],
        axis: Optional[int] = None,
        out: Optional[np.ndarray] = None,
    ) -> PassthroughTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        if out:
            return InitialGammaTensor(
                values=self._values().compress(condition, axis, out),
                entities=self._entities().compress(condition, axis, out),
                min_vals=self._min_values().compress(condition, axis, out),
                max_vals=self._max_values().compress(condition, axis, out),
                scalar_manager=self.scalar_manager,  # TODO: Check if removing this creates a backdoor to data
            )
        else:
            # TODO: Check if "out" needs to be returned at all
            out = InitialGammaTensor(
                values=self._values().compress(condition, axis, out),
                entities=self._entities().compress(condition, axis, out),
                min_vals=self._min_values().compress(condition, axis, out),
                max_vals=self._max_values().compress(condition, axis, out),
                scalar_manager=self.scalar_manager,
            )
            return out

    def __and__(
        self, other: Union[np.ndarray, IntermediateGammaTensor]
    ) -> IntermediateGammaTensor:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals and other,
                    max_vals=np.ones_like(vals),
                    min_vals=np.zeros_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                tensor = InitialGammaTensor(
                    values=self_vals and other_vals,
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    def __or__(
        self, other: Union[np.ndarray, IntermediateGammaTensor]
    ) -> IntermediateGammaTensor:
        if isinstance(other, np.ndarray):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                vals = self._values()
                tensor = InitialGammaTensor(
                    values=vals or other,
                    max_vals=np.ones_like(vals),
                    min_vals=np.zeros_like(vals),
                    entities=self._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        elif isinstance(other, IntermediateGammaTensor):
            if is_broadcastable(self.shape, other.shape):
                # relative
                from .initial_gamma import InitialGammaTensor

                self_vals = self._values()
                other_vals = other._values()
                tensor = InitialGammaTensor(
                    values=self_vals or other_vals,
                    min_vals=np.zeros_like(self_vals),
                    max_vals=np.ones_like(self_vals),
                    entities=self._entities() + other._entities(),
                )
            else:
                raise Exception(
                    f"Tensor shapes not compatible: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return tensor

    def take(
        self,
        indices: Optional[Union[int, Tuple[int, ...]]] = None,
        axis: Optional[int] = None,
        mode: str = "raise",
    ) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        if not axis:
            return InitialGammaTensor(
                values=self._values().take(indices, mode),
                entities=self._entities().take(indices, mode),
                min_vals=self._min_values().take(indices, mode),
                max_vals=self._max_values().take(indices, mode),
                scalar_manager=self.scalar_manager,
            )
        else:
            return InitialGammaTensor(
                values=self._values().take(indices, axis, mode),
                entities=self._entities().take(indices, axis, mode),
                min_vals=self._min_values().take(indices, axis, mode),
                max_vals=self._max_values().take(indices, axis, mode),
                scalar_manager=self.scalar_manager,
            )

    def diagonal(
        self, offset: int = 0, axis1: int = 0, axis2: int = 1
    ) -> IntermediateGammaTensor:
        # Note: Currently NumPy returns a read only view, but plans to return a full copy in the future
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().diagonal(offset, axis1, axis2),
            entities=self._entities().diagonal(offset, axis1, axis2),
            min_vals=self._min_values().diagonal(offset, axis1, axis2),
            max_vals=self._max_values().diagonal(offset, axis1, axis2),
            scalar_manager=self.scalar_manager,
        )

    def put(
        self,
        indices: Sequence[int],  # Union[int, Tuple[int, ...], np.ndarray],
        values: Sequence[int],  # Union[int, Tuple[int, ...], np.ndarray],
        mode: Optional[str] = "raise",
    ) -> None:
        # relative
        from .initial_gamma import InitialGammaTensor

        new_values = self._values()

        # TODO: Check what happens with entities here, if data is replaced with public values?

        if isinstance(values, np.ndarray):
            for index, value in zip(indices, values):
                new_values[index] = value

            _ = InitialGammaTensor(
                values=new_values,
                entities=self._entities(),
                min_vals=self._min_values(),
                max_vals=self._max_values(),
                scalar_manager=self.scalar_manager,
            )

        else:
            raise NotImplementedError

    def trace(
        self, offset: int = 0, axis1: int = 0, axis2: int = 1
    ) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().trace(offset, axis1, axis2),
            entities=self._entities().trace(offset, axis1, axis2),
            max_vals=self._max_values().trace(offset, axis1, axis2),
            min_vals=self._min_values().trace(offset, axis1, axis2),
            scalar_manager=self.scalar_manager,
        )

    def any(self) -> bool:
        for i in self._values():
            if i.any():
                return True
        return False

    def all(self) -> bool:
        for i in self._values():
            if not i.all():
                return False
        return True

    def __abs__(self) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=abs(self._values()),
            entities=self._entities(),
            max_vals=abs(self._max_values()),
            min_vals=abs(self._min_values()),
            scalar_manager=self.scalar_manager,
        )

    # def __divmod__(self, other: Union[int, np.ndarray]) -> IntermediateGammaTensor:
    #     # relative
    #     from .initial_gamma import InitialGammaTensor
    #
    #     return InitialGammaTensor(
    #         values=self._values() % other,
    #         entities=self._entities(),
    #         max_vals=self._max_values() % other,
    #         min_vals=self._min_values() % other,
    #     )

    def __floordiv__(self, other: Union[int, np.ndarray]) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values() // other,
            entities=self._entities(),
            max_vals=self._max_values() % other,
            min_vals=self._min_values() % other,
        )

    def cumsum(self, axis: Optional[int] = None) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().cumsum(axis),
            entities=self._entities().cumsum(axis),
            max_vals=self._max_values().cumsum(axis),
            min_vals=self._min_values().cumsum(axis),
            scalar_manager=self.scalar_manager,
        )

    def cumprod(self, axis: Optional[int] = None) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        return InitialGammaTensor(
            values=self._values().cumprod(axis),
            entities=self._entities().cumsum(
                axis
            ),  # entities get summed (combined), not multiplied
            max_vals=self._max_values().cumprod(axis),
            min_vals=self._min_values().cumprod(axis),
            scalar_manager=self.scalar_manager,
        )

    #
    # def __round__(self, n: Optional[int] = None) -> IntermediateGammaTensor:
    #     # relative
    #     from .initial_gamma import InitialGammaTensor
    #
    #     return InitialGammaTensor(
    #         values=self._values().__round__(n),
    #         entities=self._entities(),
    #         max_vals=self._max_values().__round__(n),
    #         min_vals=self._min_values().__round__(n),
    #     )

    def max(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        output_values = self._values().max(axis)
        indices = output_values.argmax(axis)

        # If indices returns a single value, then take simply returns an Entity or DSG instead of an array object
        # with an Entity/DSG inside. We will fix this by manually casting the results into arrays.

        output_entities = np.asarray(self._entities().take(indices))
        output_min = np.asarray(self._min_values().take(indices))
        output_max = np.asarray(self._max_values().take(indices))

        # TODO: Investigate if this can technically return a SEPT, if the max value only has 1 entity
        return InitialGammaTensor(
            values=output_values,
            entities=output_entities,
            max_vals=output_max,
            min_vals=output_min,
            scalar_manager=self.scalar_manager,
        )

    def min(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> IntermediateGammaTensor:
        # relative
        from .initial_gamma import InitialGammaTensor

        output_values = self._values().min(axis)
        indices = output_values.argmin(axis)

        # TODO: Investigate if this can technically return a SEPT, if the max value only has 1 entity
        return InitialGammaTensor(
            values=output_values,
            entities=np.asarray(self._entities().take(indices)),
            max_vals=np.asarray(self._max_values().take(indices)),
            min_vals=np.asarray(self._min_values().take(indices)),
            scalar_manager=self.scalar_manager,
        )



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict as TypeDict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
import uuid

# third party
import numpy as np

# relative
from .. import autograd
from ....lib.python.collections.collections import DefaultDict
from ....lib.python.collections.collections import SerializableCounter
from ...common.serde.serializable import serializable
from ..ancestors import AutogradTensorAncestor
from ..ancestors import PhiTensorAncestor
from ..passthrough import AcceptableSimpleType  # type: ignore
from ..passthrough import PassthroughTensor  # type: ignore
from ..passthrough import is_acceptable_simple_type  # type: ignore


@serializable(recursive_serde=True)
class AutogradTensor(PassthroughTensor, PhiTensorAncestor):

    __attr_allowlist__ = [
        "child",
        "requires_grad",
        "_grad",
        "_grad_fn",
        "ops",
        "backprop_id",
        "n_backwards",
    ]

    def __init__(
        self,
        child: Union[Type[AutogradTensor], AcceptableSimpleType],
        requires_grad: bool = False,
    ) -> None:
        super().__init__(child)

        # whether to run backpropagation or not
        self.requires_grad = requires_grad

        # tensor gradient
        self._grad: TypeDict = DefaultDict(lambda: None)

        # operation used to create this tensor (if any)
        self._grad_fn: Optional[Type[autograd.backward_ops.Op]] = None

        # list of ops which use this tensor
        self.ops: List = list()

        self.backprop_id: Optional[uuid.UUID] = None

        self.n_backwards: SerializableCounter = (
            SerializableCounter()
        )  # may have to add [uuid.UUID] for type annotation

    @property
    def grad(self) -> Optional[np.ndarray]:
        if self.backprop_id not in self._grad:
            return None
        return self._grad[self.backprop_id]

    @property
    def grad_fn(
        self,
    ) -> Optional[Type[autograd.backward_ops.Op]]:
        if not self.requires_grad:
            raise Exception("This tensor is not backpropagated")
        return self._grad_fn

    # Autograd Tensor Operations
    """ Note: Ignoring return type incompatibilities since AutogradTensorAncestor doesn't inherit from
    PassThroughTensor"""

    def __abs__(self) -> AutogradTensorAncestor:
        op = autograd.backward_ops.AbsOp()
        return op(self)

    def __add__(self, other: AutogradTensor) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.AddOp()
        return op(self, other)

    def __sub__(self, other: AutogradTensor) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.SubOp()
        return op(self, other)

    def __mul__(self, other: AutogradTensor) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.MulOp()
        return op(self, other)

    def __rmul__(self, other: AutogradTensor) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.MulOp()
        return op(self, other)

    def __truediv__(self, other: AutogradTensor) -> AutogradTensorAncestor:  # type: ignore
        if is_acceptable_simple_type(other):
            # Ignoring type annotation error because only int, floats, np.ndarrays will be parsed
            return self * (1 / other)  # type: ignore
        return NotImplemented

    def __pow__(self, other: Any) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.PowOp()
        return op(self, other)

    def __rpow__(self, other: Any) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.RPowOp()
        return op(self, other)

    def reshape(self, *shape: Tuple[int]) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.ReshapeOp()
        return op(self, *shape)

    def repeat(self, *args: Tuple[Any, ...], **kwargs: Any) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.RepeatOp()
        return op(self, *args, **kwargs)

    def copy(self) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.CopyOp()
        return op(self)

    def sum(self, *args: int, **kwargs: int) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.SumOp()
        return op(self, *args, **kwargs)

    def transpose(self, *dims: tuple) -> AutogradTensorAncestor:  # type: ignore
        op = autograd.backward_ops.TransposeOp()
        return op(self, *dims)

    # End Autograd Tensor Operations

    def add_grad(self, grad: np.ndarray) -> None:

        # print("Adding grad:" + str(type(grad)) + " w/ backprop_id:" + str(self.backprop_id))

        if self._grad[self.backprop_id] is None:
            self._grad[self.backprop_id] = grad
        else:

            self._grad[self.backprop_id] = self._grad[self.backprop_id] + grad

    def backward(
        self,
        grad: Optional[np.ndarray] = None,
        backprop_id: Optional[uuid.UUID] = None,
    ) -> bool:

        if backprop_id is None:
            backprop_id = uuid.uuid4()

        self.n_backwards[backprop_id] += 1

        self.backprop_id = backprop_id

        if not self.grad_fn:
            return False

        if grad is None and self._grad[self.backprop_id] is None:
            # in case if this is last loss tensor
            grad = np.ones(self.shape)
            # grad = self.__class__(grad, requires_grad=False)
            # this more or less ensures it has the right tensor chain
            # grad = (self * 0) + 1

        elif self.grad is not None:
            grad = self._grad[self.backprop_id]

        if not self.requires_grad:
            raise Exception("This tensor is not backpropagated")

        # if all gradients are accounted for - backprop
        if self.n_backwards[backprop_id] >= len(self.ops):

            self.grad_fn.backward(grad, backprop_id=backprop_id)  # type: ignore

        # if some gradietns appear to be missing - parse forward in
        # the graph to double check
        else:

            # investigate whether any of the missing ops are actually
            # going to get used.
            found_id = False

            n_direct_ops = 0
            for op in self.ops:
                if op.backprop_id is not None and op.backprop_id == backprop_id:
                    n_direct_ops += 1

            # if the number of operations we know will be backpropagating gradients to us
            # exceeds the number of times we've been backpropgated into - then we know
            # we need to wait.
            if n_direct_ops > self.n_backwards[backprop_id]:
                found_id = True

            else:

                for op in self.ops:
                    if op.backprop_id is None:
                        if op.out.find_backprop_id(self.backprop_id):
                            found_id = True
                            break

            if found_id:
                "do nothing - we're going to get another gradient"
            else:
                # backprop anyway - we've got all the grads we're gonna get
                self.grad_fn.backward(grad, backprop_id=backprop_id)  # type: ignore

        return True

    def find_backprop_id(self, backprop_id: Optional[uuid.UUID]) -> bool:
        found_id = False

        for op in self.ops:
            if op.backprop_id is not None and op.backprop_id == backprop_id:
                return True

            if op.out.find_backprop_id(self.backprop_id):
                found_id = True
                break

        return found_id



# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from .. import python as py
from ...core.common.serde.serializable import serializable
from ...core.common.uid import UID
from ...logger import traceback_and_raise
from ...proto.lib.python.iterator_pb2 import Iterator as Iterator_PB
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet


@serializable()
class Iterator(PyPrimitive):
    def __init__(self, _ref: Any, max_len: Optional[int] = None):
        super().__init__()
        self._obj_ref = _ref
        self._index = 0
        self._id = UID()
        self.max_len = max_len
        self.exhausted = False

    def __iter__(self) -> "Iterator":
        return self

    def __len__(self) -> int:
        try:
            return len(self._obj_ref)
        except Exception as e:
            traceback_and_raise(e)

    def __reduce__(self) -> Any:
        # see these tests: test_valuesiterator_pickling and test_iterator_pickling
        raise TypeError(f"Pickling {type(self)} is not supported.")

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        if hasattr(other, "_obj_ref"):
            res = self._obj_ref == other._obj_ref
        else:
            res = self._obj_ref == other

        return PrimitiveFactory.generate_primitive(value=res)

    def __next__(self) -> Any:
        # we need to do lots of getattr / setattr because some times the __next__
        # method gets called with a generator
        try:
            if hasattr(self, "_obj_ref"):
                _obj_ref = self._obj_ref
            else:
                # we got handed a generator directly into __next__
                # happens in test_reversed_iterator
                _obj_ref = self

            # max_len == None means the _ref could update while iterating. While that
            # shouldn't happen with a IteratorPointer, it can happen on a local Iterator.
            # If thats the case we just calculate it each time. Importantly we need to
            # still set exhausted otherwise the test case in list_test.py wont pass.
            max_len = None
            if hasattr(self, "max_len"):
                max_len = self.max_len

            if max_len is None:
                try:
                    if hasattr(_obj_ref, "__len__"):
                        max_len = _obj_ref.__len__()
                except AttributeError:
                    # I am not sure why this happens on some types
                    pass

            exhausted = getattr(self, "exhausted", False)
            self_index = getattr(self, "_index", 0)
            if (max_len is not None and self_index >= max_len) or exhausted:
                self.exhausted = True
                raise StopIteration

            try:
                if hasattr(_obj_ref, "__next__"):
                    try:
                        obj = next(_obj_ref)
                    except Exception as e:
                        if type(e) is StopIteration:
                            raise e
                        if type(e) is AttributeError:
                            # no _mapping exhausted?
                            raise StopIteration()
                        if type(e) is NameError:
                            # free after use?
                            raise StopIteration()

                        # test_dictitems_contains_use_after_free wants us to StopIteration
                        # test_merge_and_mutate and test_mutating_iteration wants us to
                        # raise a RuntimeError
                        # see:
                        # def test_dictitems_contains_use_after_free(self):
                        # Lets RuntimeError for now
                        raise RuntimeError

                elif hasattr(_obj_ref, "__getitem__") and hasattr(self, "_index"):
                    obj = _obj_ref[self._index]
                elif hasattr(_obj_ref, "__iter__"):
                    # collections.abc.* KeysView, ValuesView, ItemsView end up here
                    # they do not have __next__ or __getitem__ but they do have __iter__
                    # so we can just replace our self._obj_ref and keep going
                    self._obj_ref = _obj_ref.__iter__()
                    # obj = next(self._obj_ref) # just call self.__next__() instead
                    return self.__next__()
                else:
                    raise ValueError("Can't iterate through given object.")
            except StopIteration as e:
                self.exhausted = True
                raise e

            if hasattr(self, "_index"):
                self._index += 1
            return obj
        except Exception as e:
            raise e

    def upcast(self) -> Any:
        return iter(self._obj_ref)

    # TODO: Fix based on message from Tudor Cebere
    # So, when we add a new builtin type we want to have feature parity with cython ones.
    # When we tried to do this for iterators in the early days we had some problems when the iterators are infinite
    # (most likely an iterator from a generator). This pattern is common in functional programming, when you use
    # infinite iterators for different purposes. I then said that it makes sense to force the user to exhaust the
    # iterator himself and then to serde the type. Here, it might be a bit problematic because somebody might slip
    # in this kind of iterator and when we exhaust it (through list conversion), we go into infinite computation.
    # And there are similar edge cases to this.

    def _object2proto(self) -> Iterator_PB:
        id_ = sy.serialize(obj=self._id)
        obj_ref_ = sy.serialize(py.list.List(list(self._obj_ref)), to_bytes=True)
        index_ = self._index
        max_len_ = self.max_len
        exhausted_ = self.exhausted
        return Iterator_PB(
            id=id_,
            obj_ref=obj_ref_,
            index=index_,
            max_len=max_len_,
            exhausted=exhausted_,
        )

    @staticmethod
    def _proto2object(proto: Iterator_PB) -> "Iterator":
        id_: UID = sy.deserialize(blob=proto.id)
        obj_ref_ = sy.deserialize(blob=proto.obj_ref, from_bytes=True)
        index_ = proto.index
        max_len_ = proto.max_len
        exhausted_ = proto.exhausted

        new_iter = Iterator(_ref=obj_ref_, max_len=max_len_)

        new_iter._index = index_
        new_iter.exhausted = exhausted_
        new_iter._id = id_

        return new_iter

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Iterator_PB



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/store/dataset.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.store import (
    store_object_pb2 as proto_dot_core_dot_store_dot_store__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1eproto/core/store/dataset.proto\x12\x0fsyft.core.store\x1a%proto/core/common/common_object.proto\x1a#proto/core/store/store_object.proto"\xe2\x01\n\x07\x44\x61taset\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x10\n\x08obj_type\x18\x02 \x01(\t\x12\x1a\n\x12schematic_qualname\x18\x03 \x01(\t\x12-\n\x04\x64\x61ta\x18\x04 \x03(\x0b\x32\x1f.syft.core.store.StorableObject\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x18\n\x10read_permissions\x18\x07 \x01(\x0c\x12\x1a\n\x12search_permissions\x18\x08 \x01(\x0c\x62\x06proto3'
)


_DATASET = DESCRIPTOR.message_types_by_name["Dataset"]
Dataset = _reflection.GeneratedProtocolMessageType(
    "Dataset",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATASET,
        "__module__": "proto.core.store.dataset_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.Dataset)
    },
)
_sym_db.RegisterMessage(Dataset)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _DATASET._serialized_start = 128
    _DATASET._serialized_end = 354
# @@protoc_insertion_point(module_scope)



# stdlib
import importlib
import sys
from types import ModuleType
from typing import Any
from typing import Any as TypeAny
from typing import Dict as TypeDict
from typing import Iterable
from typing import List as TypeList
from typing import Optional
from typing import Set as TypeSet
from typing import Tuple as TypeTuple
from typing import Union as TypeUnion
import warnings

# third party
from cachetools import cached
from cachetools.keys import hashkey
from packaging import version

# relative
from ..ast.globals import Globals
from ..core.adp import create_adp_ast
from ..core.node.abstract.node import AbstractNodeClient
from ..core.node.common import create_client_ast
from ..core.smpc import create_smpc_ast
from ..core.tensor import create_tensor_ast
from ..logger import critical
from ..logger import traceback_and_raise
from ..logger import warning
from .misc import create_union_ast
from .numpy import create_ast
from .python import create_python_ast
from .torch import create_torch_ast


class VendorLibraryImportException(Exception):
    pass


def vendor_requirements_available(vendor_requirements: TypeDict[str, TypeAny]) -> bool:
    """
    Check whether torch or python version is supported

    Args:
        vendor_requirements: dictionary containing version of python or torch to be supported

    Returns:
        True if system supports all vendor requirements

    """
    # see if python version is supported
    if "python" in vendor_requirements:
        python_reqs = vendor_requirements["python"]

        PYTHON_VERSION = sys.version_info
        min_version = python_reqs.get("min_version", None)
        if min_version is not None:
            if PYTHON_VERSION < min_version:
                traceback_and_raise(
                    VendorLibraryImportException(
                        f"Unable to load {vendor_requirements['lib']}."
                        + f"Python: {PYTHON_VERSION} < {min_version}"
                    )
                )
        max_version = python_reqs.get("max_version", None)
        if max_version is not None:
            if PYTHON_VERSION > max_version:
                traceback_and_raise(
                    VendorLibraryImportException(
                        f"Unable to load {vendor_requirements['lib']}."
                        + f"Python: {PYTHON_VERSION} > {max_version}"
                    )
                )

    # see if torch version is supported
    if "torch" in vendor_requirements:
        torch_reqs = vendor_requirements["torch"]
        # third party
        import torch

        TORCH_VERSION = version.parse(torch.__version__.split("+")[0])
        min_version = torch_reqs.get("min_version", None)
        if min_version is not None:
            if TORCH_VERSION < version.parse(min_version):
                traceback_and_raise(
                    VendorLibraryImportException(
                        f"Unable to load {vendor_requirements['lib']}."
                        + f"Torch: {TORCH_VERSION} < {min_version}"
                    )
                )

        max_version = torch_reqs.get("max_version", None)
        if max_version is not None:
            if TORCH_VERSION > version.parse(max_version):
                traceback_and_raise(
                    VendorLibraryImportException(
                        f"Unable to load {vendor_requirements['lib']}."
                        + f"Torch: {TORCH_VERSION} > {max_version}"
                    )
                )

    return True


def _add_lib(
    *, vendor_ast: ModuleType, ast_or_client: TypeUnion[Globals, AbstractNodeClient]
) -> None:
    update_ast = getattr(vendor_ast, "update_ast", None)
    if update_ast is not None:
        update_ast(ast_or_client=ast_or_client)


def _regenerate_unions(*, lib_ast: Globals, client: TypeAny = None) -> None:
    union_misc_ast = getattr(
        getattr(create_union_ast(lib_ast=lib_ast, client=client), "syft", None),
        "lib",
        None,
    )
    if union_misc_ast is None:
        return
    elif client is not None:
        client.syft.lib.add_attr(attr_name="misc", attr=union_misc_ast.attrs["misc"])
    else:
        lib_ast.syft.lib.add_attr(attr_name="misc", attr=union_misc_ast.attrs["misc"])


def get_cache() -> TypeDict:
    return dict()


@cached(get_cache(), lambda *, lib, options=None: hashkey(lib))
def _load_lib(*, lib: str, options: Optional[TypeDict[str, TypeAny]] = None) -> None:
    """
    Load and Update Node with given library module

    Args:
        lib: name of library to load and update Node with
        options: external requirements for loading library successfully
    """
    global lib_ast
    _options = {} if options is None else options

    _ = importlib.import_module(lib)
    vendor_ast = importlib.import_module(f"syft.lib.{lib}")
    PACKAGE_SUPPORT = getattr(vendor_ast, "PACKAGE_SUPPORT", None)
    if PACKAGE_SUPPORT is None:
        raise Exception(f"Unable to load package: {lib}. Missing PACKAGE_SUPPORT.")
    PACKAGE_SUPPORT.update(_options)
    if PACKAGE_SUPPORT is not None and vendor_requirements_available(
        vendor_requirements=PACKAGE_SUPPORT
    ):
        _add_lib(vendor_ast=vendor_ast, ast_or_client=lib_ast)
        # cache the constructor for future created clients
        update_ast_func = getattr(vendor_ast, "update_ast", None)
        if update_ast_func is None:
            raise Exception(f"Unable to load package: {lib}. Missing update_ast func")
        lib_ast.loaded_lib_constructors[lib] = update_ast_func
        _regenerate_unions(lib_ast=lib_ast)

        for _, client in lib_ast.registered_clients.items():
            _add_lib(vendor_ast=vendor_ast, ast_or_client=client)
            _regenerate_unions(lib_ast=lib_ast, client=client)


def load(
    *libs: TypeUnion[TypeList[str], TypeTuple[str], TypeSet[str], str],
    options: Optional[TypeDict[str, TypeAny]] = None,
    ignore_warning: bool = False,
    **kwargs: str,
) -> None:
    """
    Load and Update Node with given library module

    Args:
        *libs: names of libraries to load and update Node with (can be variadic, tuple, list, set)
        options: external requirements for loading library successfully
        **kwargs: for backward compatibility with calls like `syft.load(lib = "opacus")`
    """
    # For backward compatibility with calls like `syft.load(lib = "opacus")`
    # Note: syft.load(lib = "opacus") doesnot work as it iterates the string, syft.load('opacus') works
    options = options if options is not None else {}

    if not ignore_warning:
        msg = "sy.load() is deprecated and not needed anymore"
        warning(msg, print=True)
        warnings.warn(msg, DeprecationWarning)

    if "lib" in kwargs.keys():
        libs += tuple(kwargs["lib"])

    if isinstance(libs[0], Iterable):
        if not isinstance(libs[0], str):
            libs = tuple(libs[0])
        for lib in libs:
            if isinstance(lib, str):
                try:
                    _load_lib(lib=str(lib), options=options)
                except VendorLibraryImportException as e:
                    warning(e)
                except Exception as e:
                    warning(f"Unable to load package support for: {lib}. {e}")
            else:
                warning(
                    f"Unable to load package support for: {lib}. Pass lib name as string object."
                )
    else:
        critical(
            "Unable to load package support for any library. Iterable object not found."
        )


def load_lib(lib: str, options: Optional[TypeDict[str, TypeAny]] = None) -> None:
    """
    Load and Update Node with given library module
    _load_lib() is deprecated please use load() in the future

    Args:
        lib: name of library to load and update Node with
        options: external requirements for loading library successfully

    """
    options = options if options is not None else {}
    msg = "sy._load_lib() is deprecated and not needed anymore"
    warning(msg, print=True)
    warnings.warn(msg, DeprecationWarning)
    load(lib=lib, options=options)


# now we need to load the relevant frameworks onto the node
def create_lib_ast(client: Optional[Any] = None) -> Globals:
    """
    Create AST and load the relevant frameworks onto the node

    Args:
        client: VM client onto whom the frameworks need to be loaded

    Returns:
        AST for client of type Globals

    """
    python_ast = create_python_ast(client=client)
    torch_ast = create_torch_ast(client=client)
    numpy_ast = create_ast(client=client)
    adp_ast = create_adp_ast(client=client)
    client_ast = create_client_ast(client=client)
    tensor_ast = create_tensor_ast(client=client)
    smpc_ast = create_smpc_ast(client=client)

    lib_ast = Globals(client=client)
    lib_ast.add_attr(attr_name="syft", attr=python_ast.attrs["syft"])
    lib_ast.add_attr(attr_name="torch", attr=torch_ast.attrs["torch"])
    lib_ast.add_attr(attr_name="numpy", attr=numpy_ast.attrs["numpy"])
    lib_ast.syft.add_attr("core", attr=adp_ast.syft.core)
    lib_ast.syft.core.add_attr("node", attr=client_ast.syft.core.node)
    lib_ast.syft.core.add_attr("common", attr=client_ast.syft.core.common)
    lib_ast.syft.core.add_attr("tensor", attr=tensor_ast.syft.core.tensor)
    lib_ast.syft.core.add_attr("smpc", attr=smpc_ast.syft.core.smpc)

    # let the misc creation be always the last, as it needs the full ast solved
    # to properly generated unions
    union_misc_ast = getattr(
        getattr(create_union_ast(lib_ast, client), "syft", None), "lib", None
    )
    if union_misc_ast:
        lib_ast.syft.lib.add_attr(attr_name="misc", attr=union_misc_ast.attrs["misc"])

    return lib_ast


lib_ast = create_lib_ast(None)


# @wrapt.when_imported("numpy")
# def post_import_hook_third_party(module: TypeAny) -> None:
#     """
#     Note: This needs to be after `lib_ast` because code above uses lib-ast
#     """
#     # msg = f"inside post_import_hook_third_party module_name {module.__name__}"
#     # warning(msg, print=True)
#     # warnings.warn(msg, DeprecationWarning)
#     load(module.__name__, ignore_warning=True)



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

# third party
import numpy as np

# relative
from .tensor import Tensor
from .util import implements
from .util import inputs2child


@implements(Tensor, np.mean)
def mean(
    array: np.typing.ArrayLike, axis: Optional[int] = None, **kwargs: Any
) -> Tensor:
    if axis is None:
        den = float(np.prod(array.shape))
    else:
        den = array.shape[axis]

    return array.sum(axis=axis) / den


@implements(Tensor, np.max)
def npmax(*args: Tuple[Any, ...], **kwargs: Dict[Any, Any]) -> Tensor:
    args, kwargs = inputs2child(*args, **kwargs)  # type: ignore
    return np.max(*args, **kwargs)


@implements(Tensor, np.min)
def npmin(*args: List[Any], **kwargs: Dict[Any, Any]) -> Tensor:
    args, kwargs = inputs2child(*args, **kwargs)  # type: ignore
    return np.min(*args, **kwargs)


@implements(Tensor, np.square)
def square(x: Tensor) -> Tensor:
    return x * x


@implements(Tensor, np.expand_dims)
def expand_dims(*args: List[Any], **kwargs: Dict[Any, Any]) -> Tensor:
    args, kwargs = inputs2child(*args, **kwargs)  # type: ignore
    return Tensor(np.expand_dims(*args, **kwargs))


@implements(Tensor, np.multiply)
def multiply(a: Tensor, b: Tensor) -> Tensor:
    if isinstance(a, Tensor):
        result = a.__mul__(b)
        if result is not NotImplementedError:
            return result

    if isinstance(b, Tensor):
        result = b.__rmul__(a)
        if result is not NotImplementedError:
            return result

    raise TypeError(f"Can't multiply {type(a)} with {type(b)}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/dataset_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*proto/grid/messages/dataset_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xac\x02\n\x14\x43reateDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x64\x61taset\x18\x03 \x01(\x0c\x12H\n\x08metadata\x18\x04 \x03(\x0b\x32\x36.syft.grid.messages.CreateDatasetMessage.MetadataEntry\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x10\n\x08platform\x18\x06 \x01(\t\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x9f\x01\n\x11GetDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\ndataset_id\x18\x03 \x01(\x03\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xdc\x01\n\x12GetDatasetResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x46\n\x08metadata\x18\x02 \x03(\x0b\x32\x34.syft.grid.messages.GetDatasetResponse.MetadataEntry\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x8c\x01\n\x12GetDatasetsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\xd7\x02\n\x13GetDatasetsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12M\n\tmetadatas\x18\x02 \x03(\x0b\x32:.syft.grid.messages.GetDatasetsResponse.metadata_container\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x1a\xa1\x01\n\x12metadata_container\x12Z\n\x08metadata\x18\x01 \x03(\x0b\x32H.syft.grid.messages.GetDatasetsResponse.metadata_container.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01"\x9d\x02\n\x14UpdateDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\ndataset_id\x18\x03 \x01(\x03\x12H\n\x08metadata\x18\x04 \x03(\x0b\x32\x36.syft.grid.messages.UpdateDatasetMessage.MetadataEntry\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xb9\x01\n\x14\x44\x65leteDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\ndataset_id\x18\x03 \x01(\t\x12\x15\n\rbin_object_id\x18\x04 \x01(\t\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_CREATEDATASETMESSAGE = DESCRIPTOR.message_types_by_name["CreateDatasetMessage"]
_CREATEDATASETMESSAGE_METADATAENTRY = _CREATEDATASETMESSAGE.nested_types_by_name[
    "MetadataEntry"
]
_GETDATASETMESSAGE = DESCRIPTOR.message_types_by_name["GetDatasetMessage"]
_GETDATASETRESPONSE = DESCRIPTOR.message_types_by_name["GetDatasetResponse"]
_GETDATASETRESPONSE_METADATAENTRY = _GETDATASETRESPONSE.nested_types_by_name[
    "MetadataEntry"
]
_GETDATASETSMESSAGE = DESCRIPTOR.message_types_by_name["GetDatasetsMessage"]
_GETDATASETSRESPONSE = DESCRIPTOR.message_types_by_name["GetDatasetsResponse"]
_GETDATASETSRESPONSE_METADATA_CONTAINER = _GETDATASETSRESPONSE.nested_types_by_name[
    "metadata_container"
]
_GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY = (
    _GETDATASETSRESPONSE_METADATA_CONTAINER.nested_types_by_name["MetadataEntry"]
)
_UPDATEDATASETMESSAGE = DESCRIPTOR.message_types_by_name["UpdateDatasetMessage"]
_UPDATEDATASETMESSAGE_METADATAENTRY = _UPDATEDATASETMESSAGE.nested_types_by_name[
    "MetadataEntry"
]
_DELETEDATASETMESSAGE = DESCRIPTOR.message_types_by_name["DeleteDatasetMessage"]
CreateDatasetMessage = _reflection.GeneratedProtocolMessageType(
    "CreateDatasetMessage",
    (_message.Message,),
    {
        "MetadataEntry": _reflection.GeneratedProtocolMessageType(
            "MetadataEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _CREATEDATASETMESSAGE_METADATAENTRY,
                "__module__": "proto.grid.messages.dataset_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateDatasetMessage.MetadataEntry)
            },
        ),
        "DESCRIPTOR": _CREATEDATASETMESSAGE,
        "__module__": "proto.grid.messages.dataset_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateDatasetMessage)
    },
)
_sym_db.RegisterMessage(CreateDatasetMessage)
_sym_db.RegisterMessage(CreateDatasetMessage.MetadataEntry)

GetDatasetMessage = _reflection.GeneratedProtocolMessageType(
    "GetDatasetMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETDATASETMESSAGE,
        "__module__": "proto.grid.messages.dataset_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetDatasetMessage)
    },
)
_sym_db.RegisterMessage(GetDatasetMessage)

GetDatasetResponse = _reflection.GeneratedProtocolMessageType(
    "GetDatasetResponse",
    (_message.Message,),
    {
        "MetadataEntry": _reflection.GeneratedProtocolMessageType(
            "MetadataEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _GETDATASETRESPONSE_METADATAENTRY,
                "__module__": "proto.grid.messages.dataset_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.GetDatasetResponse.MetadataEntry)
            },
        ),
        "DESCRIPTOR": _GETDATASETRESPONSE,
        "__module__": "proto.grid.messages.dataset_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetDatasetResponse)
    },
)
_sym_db.RegisterMessage(GetDatasetResponse)
_sym_db.RegisterMessage(GetDatasetResponse.MetadataEntry)

GetDatasetsMessage = _reflection.GeneratedProtocolMessageType(
    "GetDatasetsMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETDATASETSMESSAGE,
        "__module__": "proto.grid.messages.dataset_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetDatasetsMessage)
    },
)
_sym_db.RegisterMessage(GetDatasetsMessage)

GetDatasetsResponse = _reflection.GeneratedProtocolMessageType(
    "GetDatasetsResponse",
    (_message.Message,),
    {
        "metadata_container": _reflection.GeneratedProtocolMessageType(
            "metadata_container",
            (_message.Message,),
            {
                "MetadataEntry": _reflection.GeneratedProtocolMessageType(
                    "MetadataEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY,
                        "__module__": "proto.grid.messages.dataset_messages_pb2"
                        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetDatasetsResponse.metadata_container.MetadataEntry)
                    },
                ),
                "DESCRIPTOR": _GETDATASETSRESPONSE_METADATA_CONTAINER,
                "__module__": "proto.grid.messages.dataset_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.GetDatasetsResponse.metadata_container)
            },
        ),
        "DESCRIPTOR": _GETDATASETSRESPONSE,
        "__module__": "proto.grid.messages.dataset_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetDatasetsResponse)
    },
)
_sym_db.RegisterMessage(GetDatasetsResponse)
_sym_db.RegisterMessage(GetDatasetsResponse.metadata_container)
_sym_db.RegisterMessage(GetDatasetsResponse.metadata_container.MetadataEntry)

UpdateDatasetMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateDatasetMessage",
    (_message.Message,),
    {
        "MetadataEntry": _reflection.GeneratedProtocolMessageType(
            "MetadataEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _UPDATEDATASETMESSAGE_METADATAENTRY,
                "__module__": "proto.grid.messages.dataset_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateDatasetMessage.MetadataEntry)
            },
        ),
        "DESCRIPTOR": _UPDATEDATASETMESSAGE,
        "__module__": "proto.grid.messages.dataset_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateDatasetMessage)
    },
)
_sym_db.RegisterMessage(UpdateDatasetMessage)
_sym_db.RegisterMessage(UpdateDatasetMessage.MetadataEntry)

DeleteDatasetMessage = _reflection.GeneratedProtocolMessageType(
    "DeleteDatasetMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEDATASETMESSAGE,
        "__module__": "proto.grid.messages.dataset_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteDatasetMessage)
    },
)
_sym_db.RegisterMessage(DeleteDatasetMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEDATASETMESSAGE_METADATAENTRY._options = None
    _CREATEDATASETMESSAGE_METADATAENTRY._serialized_options = b"8\001"
    _GETDATASETRESPONSE_METADATAENTRY._options = None
    _GETDATASETRESPONSE_METADATAENTRY._serialized_options = b"8\001"
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._options = None
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_options = b"8\001"
    _UPDATEDATASETMESSAGE_METADATAENTRY._options = None
    _UPDATEDATASETMESSAGE_METADATAENTRY._serialized_options = b"8\001"
    _CREATEDATASETMESSAGE._serialized_start = 135
    _CREATEDATASETMESSAGE._serialized_end = 435
    _CREATEDATASETMESSAGE_METADATAENTRY._serialized_start = 388
    _CREATEDATASETMESSAGE_METADATAENTRY._serialized_end = 435
    _GETDATASETMESSAGE._serialized_start = 438
    _GETDATASETMESSAGE._serialized_end = 597
    _GETDATASETRESPONSE._serialized_start = 600
    _GETDATASETRESPONSE._serialized_end = 820
    _GETDATASETRESPONSE_METADATAENTRY._serialized_start = 388
    _GETDATASETRESPONSE_METADATAENTRY._serialized_end = 435
    _GETDATASETSMESSAGE._serialized_start = 823
    _GETDATASETSMESSAGE._serialized_end = 963
    _GETDATASETSRESPONSE._serialized_start = 966
    _GETDATASETSRESPONSE._serialized_end = 1309
    _GETDATASETSRESPONSE_METADATA_CONTAINER._serialized_start = 1148
    _GETDATASETSRESPONSE_METADATA_CONTAINER._serialized_end = 1309
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_start = 1262
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_end = 1309
    _UPDATEDATASETMESSAGE._serialized_start = 1312
    _UPDATEDATASETMESSAGE._serialized_end = 1597
    _UPDATEDATASETMESSAGE_METADATAENTRY._serialized_start = 388
    _UPDATEDATASETMESSAGE_METADATAENTRY._serialized_end = 435
    _DELETEDATASETMESSAGE._serialized_start = 1600
    _DELETEDATASETMESSAGE._serialized_end = 1785
# @@protoc_insertion_point(module_scope)



"""In this class, we support the functionality necessary to support
virtual network connections between the nodes in the Syft ecosystem.
Replacing this object with an actual network connection object
(such as one powered by P2P tech, web sockets, or HTTP) should
execute the exact same functionality but do so over a network"""

# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...proto.core.io.connection_pb2 import (
    VirtualClientConnection as VirtualClientConnection_PB,
)
from ...proto.core.io.connection_pb2 import (
    VirtualServerConnection as VirtualServerConnection_PB,
)
from ..common.message import SignedEventualSyftMessageWithoutReply
from ..common.message import SignedImmediateSyftMessageWithReply
from ..common.message import SignedImmediateSyftMessageWithoutReply
from ..common.serde.deserialize import _deserialize
from ..node.abstract.node import AbstractNode
from .connection import ClientConnection
from .connection import ServerConnection


@final
class VirtualServerConnection(ServerConnection):
    def __init__(self, node: AbstractNode):
        self.node = node

    def recv_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        return self.node.recv_immediate_msg_with_reply(msg=msg)

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        self.node.recv_immediate_msg_without_reply(msg=msg)

    def recv_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        self.node.recv_eventual_msg_without_reply(msg=msg)

    def _object2proto(self) -> VirtualServerConnection_PB:
        return VirtualServerConnection_PB(node=self.node._object2proto())

    @staticmethod
    def _proto2object(proto: VirtualServerConnection_PB) -> "VirtualServerConnection":
        node = _deserialize(blob=proto.node, from_proto=True)
        return VirtualServerConnection(
            node=node,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return VirtualServerConnection_PB


@final
class VirtualClientConnection(ClientConnection):
    def __init__(self, server: VirtualServerConnection):
        self.server = server

    def send_immediate_msg_without_reply(
        self,
        msg: SignedImmediateSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        self.server.recv_immediate_msg_without_reply(msg=msg)

    def send_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply, timeout: Optional[float] = None
    ) -> SignedImmediateSyftMessageWithoutReply:
        return self.server.recv_immediate_msg_with_reply(msg=msg)

    def send_eventual_msg_without_reply(
        self,
        msg: SignedEventualSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        return self.server.recv_eventual_msg_without_reply(msg=msg)

    def _object2proto(self) -> VirtualClientConnection_PB:
        return VirtualClientConnection_PB(server=self.server._object2proto())

    @staticmethod
    def _proto2object(proto: VirtualClientConnection_PB) -> "VirtualClientConnection":
        return VirtualClientConnection(
            server=VirtualServerConnection._proto2object(proto.server),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return VirtualClientConnection_PB


def create_virtual_connection(node: AbstractNode) -> VirtualClientConnection:

    server = VirtualServerConnection(node=node)
    client = VirtualClientConnection(server=server)

    return client



# stdlib
from uuid import UUID

# third party
from numpy import ndarray

# relative
from ..tensor import AutogradTensor
from .op import Op


class CopyOp(Op):
    """Copy a tensor"""

    def forward(self, x: AutogradTensor) -> AutogradTensor:  # type: ignore
        self.x = x

        return AutogradTensor(x.child.copy(), requires_grad=x.requires_grad)

    def _backward(self, grad: ndarray, backprop_id: UUID) -> None:

        if self.x.requires_grad:

            self.x.add_grad(grad.copy())

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)



# third party
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class DatasetGroup(Base):
    __tablename__ = "datasetgroup"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    bin_object = Column(String(), ForeignKey("bin_object.id"))
    dataset = Column(String(), ForeignKey("json_object.id"))

    def __str__(self) -> str:
        return (
            f"<DatasetGroup id: {self.id}, bin_object: {self.bin_object}, "
            f"dataset: {self.dataset}>"
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/auth/signed_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# third party
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2

# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n$proto/core/auth/signed_message.proto\x12\x0esyft.core.auth\x1a%proto/core/common/common_object.proto\x1a\x1bgoogle/protobuf/empty.proto"\x80\x01\n\rSignedMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x10\n\x08obj_type\x18\x02 \x01(\t\x12\x11\n\tsignature\x18\x03 \x01(\x0c\x12\x12\n\nverify_key\x18\x04 \x01(\x0c\x12\x0f\n\x07message\x18\x05 \x01(\x0c"\x1f\n\tVerifyKey\x12\x12\n\nverify_key\x18\x01 \x01(\x0c"0\n\tVerifyAll\x12#\n\x03\x61ll\x18\x01 \x01(\x0b\x32\x16.google.protobuf.Emptyb\x06proto3'
)


_SIGNEDMESSAGE = DESCRIPTOR.message_types_by_name["SignedMessage"]
_VERIFYKEY = DESCRIPTOR.message_types_by_name["VerifyKey"]
_VERIFYALL = DESCRIPTOR.message_types_by_name["VerifyAll"]
SignedMessage = _reflection.GeneratedProtocolMessageType(
    "SignedMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _SIGNEDMESSAGE,
        "__module__": "proto.core.auth.signed_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.auth.SignedMessage)
    },
)
_sym_db.RegisterMessage(SignedMessage)

VerifyKey = _reflection.GeneratedProtocolMessageType(
    "VerifyKey",
    (_message.Message,),
    {
        "DESCRIPTOR": _VERIFYKEY,
        "__module__": "proto.core.auth.signed_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.auth.VerifyKey)
    },
)
_sym_db.RegisterMessage(VerifyKey)

VerifyAll = _reflection.GeneratedProtocolMessageType(
    "VerifyAll",
    (_message.Message,),
    {
        "DESCRIPTOR": _VERIFYALL,
        "__module__": "proto.core.auth.signed_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.auth.VerifyAll)
    },
)
_sym_db.RegisterMessage(VerifyAll)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SIGNEDMESSAGE._serialized_start = 125
    _SIGNEDMESSAGE._serialized_end = 253
    _VERIFYKEY._serialized_start = 255
    _VERIFYKEY._serialized_end = 286
    _VERIFYALL._serialized_start = 288
    _VERIFYALL._serialized_end = 336
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/parameter.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1fproto/lib/torch/parameter.proto\x12\x0esyft.lib.torch"}\n\x0eParameterProto\x12\x0e\n\x06tensor\x18\x01 \x01(\x0c\x12\x15\n\rrequires_grad\x18\x02 \x01(\x08\x12\x11\n\x04grad\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x12\x18\n\x0bgrad_sample\x18\x04 \x01(\x0cH\x01\x88\x01\x01\x42\x07\n\x05_gradB\x0e\n\x0c_grad_sampleb\x06proto3'
)


_PARAMETERPROTO = DESCRIPTOR.message_types_by_name["ParameterProto"]
ParameterProto = _reflection.GeneratedProtocolMessageType(
    "ParameterProto",
    (_message.Message,),
    {
        "DESCRIPTOR": _PARAMETERPROTO,
        "__module__": "proto.lib.torch.parameter_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.ParameterProto)
    },
)
_sym_db.RegisterMessage(ParameterProto)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _PARAMETERPROTO._serialized_start = 51
    _PARAMETERPROTO._serialized_end = 176
# @@protoc_insertion_point(module_scope)



"""Crypto Primitives."""

# stdlib
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional


class CryptoPrimitiveProvider:
    """A trusted third party should use this class to generate crypto primitives."""

    _func_providers: Dict[str, Callable] = {}
    _ops_list: DefaultDict[str, List] = defaultdict(list)

    def __init__(self) -> None:  # noqa
        raise ValueError("This class should not be initialized")

    @staticmethod
    def generate_primitives(
        op_str: str,
        parties: List[Any],
        g_kwargs: Optional[Dict[str, Any]] = None,
        p_kwargs: Optional[Dict[str, Any]] = None,
        nr_instances: int = 1,
        ring_size: int = 2**32,
    ) -> List[Any]:
        g_kwargs = g_kwargs if g_kwargs is not None else {}
        p_kwargs = p_kwargs if p_kwargs is not None else {}
        """Generate "op_str" primitives.

        Args:
            op_str (str): Operator.
            parties (List[Any]): Parties to generate primitives for.
            g_kwargs: Generate kwargs passed to the registered function.
            p_kwargs: Populate kwargs passed to the registered populate function.

        Returns:
            List[Any]: List of primitives.

        Raises:
            ValueError: If op_str is not registered.

        """
        if op_str not in CryptoPrimitiveProvider._func_providers:
            raise ValueError(f"{op_str} not registered")

        nr_parties = len(parties)

        generator = CryptoPrimitiveProvider._func_providers[op_str]
        primitives = generator(
            **g_kwargs,
            nr_instances=nr_instances,
            nr_parties=nr_parties,
            ring_size=ring_size,
        )

        if p_kwargs is not None:
            """Do not transfer the primitives if there is not specified a
            values for populate kwargs."""
            CryptoPrimitiveProvider._transfer_primitives_to_parties(
                op_str=op_str,
                primitives=primitives,
                parties=parties,
                p_kwargs=p_kwargs,
                ring_size=ring_size,
            )

        # Since we do not have (YET!) the possiblity to return typed tuples from a remote
        # execute function we are using this.
        return primitives

    @staticmethod
    def _transfer_primitives_to_parties(
        op_str: str,
        primitives: List[Any],
        parties: List[Any],
        p_kwargs: Dict[str, Any],
        ring_size: int,
    ) -> None:
        if not isinstance(primitives, list):
            raise ValueError("Primitives should be a List")

        if len(primitives) != len(parties):
            raise ValueError(
                f"Primitives length {len(primitives)} != Parties length {len(parties)}"
            )

        for primitives_party, party in zip(primitives, parties):
            party.syft.core.tensor.smpc.share_tensor.populate_store(
                op_str, primitives_party, **p_kwargs, ring_size=ring_size
            )

    @staticmethod
    def get_state() -> str:
        """Get the state of a CryptoProvider.

        Returns:
            str: CryptoProvider
        """
        res = f"Providers: {list(CryptoPrimitiveProvider._func_providers.keys())}\n"
        return res



# stdlib
from datetime import datetime
from typing import Any

# third party
from sqlalchemy.engine import Engine

# relative
from ...domain.enums import RequestAPIFields
from ..exceptions import AssociationRequestError
from ..node_table.association_request import AssociationRequest
from .database_manager import DatabaseManager


class AssociationRequestManager(DatabaseManager):

    schema = AssociationRequest

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=AssociationRequestManager.schema, db=database)

    def first(self, **kwargs: Any) -> AssociationRequest:
        result = super().first(**kwargs)
        if not result:
            raise AssociationRequestError

        return result

    def create_association_request(
        self,
        node_name: str,
        node_address: str,
        source: str,
        target: str,
        status: str,
    ) -> None:
        table_fields = {}
        table_fields["node_name"] = node_name
        table_fields[RequestAPIFields.REQUESTED_DATE] = datetime.now().strftime(
            "%m/%d/%Y"
        )

        table_fields[RequestAPIFields.SOURCE] = source
        table_fields[RequestAPIFields.TARGET] = target
        table_fields[RequestAPIFields.STATUS] = status
        table_fields["node_address"] = node_address

        try:
            self.first(**{"source": source, "target": target})
            self.modify(query={"source": source, "target": target}, values=table_fields)
        except AssociationRequestError:
            # no existing AssociationRequest so lets make one
            self.register(**table_fields)  # type: ignore

    def set(self, source: str, target: str, response: str) -> None:
        self.modify(
            query={"source": source, "target": target},
            values={
                "status": response,
                "accepted_date": datetime.now().strftime("%m/%d/%Y"),
            },
        )



# stdlib
from abc import abstractmethod
from typing import Any
from typing import Optional
from uuid import UUID

# third party
from numpy import ndarray

# relative
from ..tensor import AutogradTensor


class Op:
    def __init__(self) -> None:
        self.backprop_id: Optional[UUID] = None

    @abstractmethod
    def forward(self, *args: AutogradTensor, **kwargs: AutogradTensor) -> Any:
        raise NotImplementedError

    def _backward(self, grad: ndarray, backprop_id: UUID) -> Any:
        raise NotImplementedError

    def backward(self, grad: ndarray, backprop_id: UUID) -> Any:

        self.backprop_id = backprop_id

        for t in self.parent_tensors:
            t.backprop_id = backprop_id

        return self._backward(grad=grad, backprop_id=backprop_id)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:

        self.parent_tensors = list()

        for arg in args:
            if isinstance(arg, AutogradTensor):
                arg.ops.append(self)
                self.parent_tensors.append(arg)

        for _, arg in kwargs.items():
            if isinstance(arg, AutogradTensor):
                arg.ops.append(self)
                self.parent_tensors.append(arg)

        self.out = self.forward(*args, **kwargs)
        self.out._grad_fn = self

        return self.out



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from .....common.message import ImmediateSyftMessageWithoutReply
from ....abstract.node import AbstractNode
from ...action.common import EventualActionWithoutReply
from ...action.common import ImmediateActionWithReply
from ...action.common import ImmediateActionWithoutReply
from ..node_service import EventualNodeServiceWithoutReply
from ..node_service import ImmediateNodeServiceWithReply
from ..node_service import ImmediateNodeServiceWithoutReply


class ImmediateObjectActionServiceWithoutReply(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: ImmediateActionWithoutReply,
        verify_key: Optional[VerifyKey] = None,
    ) -> None:
        msg.execute_action(node=node, verify_key=verify_key)

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateActionWithoutReply]]:
        return [ImmediateActionWithoutReply]


class EventualObjectActionServiceWithoutReply(EventualNodeServiceWithoutReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: EventualActionWithoutReply,
        verify_key: Optional[VerifyKey] = None,
    ) -> None:
        msg.execute_action(node=node, verify_key=verify_key)

    @staticmethod
    def message_handler_types() -> List[Type[EventualActionWithoutReply]]:
        return [EventualActionWithoutReply]


class ImmediateObjectActionServiceWithReply(ImmediateNodeServiceWithReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: ImmediateActionWithReply,
        verify_key: Optional[VerifyKey] = None,
    ) -> ImmediateSyftMessageWithoutReply:
        return msg.execute_action(node=node, verify_key=verify_key)

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateActionWithReply]]:
        return [ImmediateActionWithReply]



"""Package where needed crypto material is stored."""

# stdlib
# stdlib
from typing import Any
from typing import Callable

# relative
from .crypto_primitive_provider import CryptoPrimitiveProvider
from .crypto_store import CryptoStore


def register_primitive_generator(name: str) -> Callable[..., Any]:
    """Decorator to register a crypto primitive provider.

    Args:
        name (str): Name of the primitive.

    Returns:
        Callable[..., Any]: returns the triple generator function.
    """

    def register_generator(func_generator: Callable[..., Any]) -> Callable[..., Any]:
        if name in CryptoPrimitiveProvider._func_providers:
            raise ValueError(f"Provider {name} already in _func_providers")
        CryptoPrimitiveProvider._func_providers[name] = func_generator
        return func_generator

    return register_generator


def register_primitive_store_add(name: str) -> Callable[..., Any]:
    """Decorator to add primitives to the store.

    Args:
        name (str): Name of the primitive.

    Returns:
        Callable[..., Any]: returns the crypto store add function.
    """

    def register_add(func_add: Callable[..., Any]) -> Callable[..., Any]:
        if name in CryptoStore._func_add_store:
            raise ValueError(f"Crypto Store 'adder' {name} already in _func_add_store")
        CryptoStore._func_add_store[name] = func_add
        return func_add

    return register_add


def register_primitive_store_get(name: str) -> Callable[..., Any]:
    """Decorator to retrieve primitives from the store.

    Args:
        name (str): Name of the primitive.

    Returns:
        Callable[..., Any]: returns the crypto store get function.
    """

    def register_get(func_get: Callable[..., Any]) -> Callable[..., Any]:
        if name in CryptoStore._func_get_store:
            raise ValueError(f"Crypto Store 'getter' {name} already in _func_get_store")
        CryptoStore._func_get_store[name] = func_get
        return func_get

    return register_get


__all__ = ["CryptoStore", "CryptoPrimitiveProvider"]
